//========================================================== file = mm1.c =====
//=  A simple "straight C" M/M/1 queue simulation                             =
//=============================================================================
//=  Notes: 1) This program is adapted from Figure 1.6 in Simulating          =
//=            Computer Systems, Techniqyes and Tools by M. H. MacDougall     =
//=            (1987).                                                        =
//=         2) Choice of two RNG's is made in expntl() before compile         =
//=---------------------------------------------------------------------------=
//=  Build: bcc32 mm1.c                                                       =
//=---------------------------------------------------------------------------=
//=  Execute: mm1 ERROR TARGET_UTIL                                           =
//=---------------------------------------------------------------------------=
//=  History: KJC (03/09/99) - Genesis                                        =
//=           KJC (12/02/02) - Modified to take parameters from command line  =
//=           KJC (12/02/02) - Added Jain's RNG (rand_val())                  =
//=           KJC (12/02/02) - Added MacDougall's RNG (randf())               =
//=============================================================================

//----- Include files ---------------------------------------------------------
#include <stdio.h>             // Needed for printf()
#include <stdlib.h>            // Needed for exit() and atof()
#include <math.h>              // Needed for log()

//----- Constants -------------------------------------------------------------
#define SERV_TIME  1.00        // Mean service time
#define INFIN      1.0e100     // Infinity (or darn close)
#define MAX_ITER     1000      // Maximum number of iterations

//----- Function prototypes ---------------------------------------------------
double expntl(double x);       // Generate exponential RV with mean x
double rand_val(long seed);    // Jain RNG

//===== Main program ==========================================================
void main()
{
  double   ERRORX;               // Error percentage
  double   TARGET_UTIL;         // Target utilization in %
  double   ARR_TIME;            // Mean interarrival time

  double   Ta;                  // Mean time between arrivals
  double   Ts;                  // Mean service time

  double   time = 0.0;         // Simulation time
  double   t1 = 0.0;           // Time for next event #1 (arrival)
  double   t2;                 // Time for next event #2 (departure)
  unsigned long int n = 0;     // Number of customers in the system

  unsigned long int c = 0;     // Number of service completions
  double   b = 0.0;            // Total busy time
  double   s = 0.0;            // Area of number of customers in system
  double   tn;                 // Variable for "last event time"
  double   tb;                 // Variable for "last start of busy time"
  double   x;                  // Throughput
  double   u;                  // Utilization
  double   l;                  // Mean number in the system
  double   l_theory;           // Theoretical l value
  double   w;                  // Mean residence time

  int      i;
  double   sum;

  FILE     *st1;                 //File stream #1
  FILE     *st2;                 //File Stream #2


  // Open input file
  st1 = fopen("c:/temp/infile", "r");
  if (st1 == NULL)
  {
    printf("*** ERROR in run() - unable to open input file\n");
  }

  // Open output file
  st2 = fopen("c:/temp/outfile", "w");
  if (st2 == NULL)
  {
    printf("*** ERROR in run() - unable to open output file\n");
  }

  //Get the allowable error and targer utilization from standard input
  fscanf(st1, "%lf %lf", &ERRORX, &TARGET_UTIL);
   
  //Close input file
  fclose(st1);
  ARR_TIME = SERV_TIME / (TARGET_UTIL / 100.0);

  sum = 0;
  for (i=1; i<MAX_ITER; i++)
  {
    // Initialize stuff
    time = t1 = n = c = b = s = 0.0;

    // Call rand once to set the seed for this run
    rand_val(i);

    // Intialize l_theory
    l_theory = TARGET_UTIL / (100.0 - TARGET_UTIL);

    // Initialize key variables
    Ta = ARR_TIME;
    Ts = SERV_TIME;
    t2 = tn = INFIN;

    // Main simulation loop
    while (1)
    {
      if (t1 < t2)                 // *** Event #1 (arrival)
      {
        time = t1;
        s = s + n * (time - tn);   // Update area under "s" curve
        n++;
        tn = time;                 // tn = "last event time" for next event
        t1 = time + expntl(Ta);
        if (n == 1)
        {
          tb = time;               // Set "last start of busy time"
          t2 = time + expntl(Ts);
        }
      }
      else                         // *** Event #2 (departure)
      {
        time = t2;
        s = s + n * (time - tn);   // Update area under "s" curve
        n--;
        tn = time;                 // tn = "last event time" for next event
        c++;                       // Increment number of completions
        if (n > 0)
          t2 = time + expntl(Ts);
        else
        {
          t2 = INFIN;
          b = b + time - tb;      // Update busy time sum if empty
        }
      }

      if (time > 0.0)
      {
        l = s / time;   // Compute mean number in system
        if ((l >= l_theory * (1 - ERRORX / 100.0)) &&
            (l <= l_theory * (1 + ERRORX / 100.0)))
        {
          break;
        }
      }
    }

    x = c / time;   // Compute throughput rate
    u = b / time;   // Compute server utilization
    l = s / time;   // Compute mean number in system
    w = l / x;      // Compute mean residence or system time

    // Output results
    //printf("=============================================================== \n");
    //printf("=            *** Results from M/M/1 Simulation ***            = \n");
    //printf("=============================================================== \n");
    //printf("=  ERROR                      = %f %%      \n", ERRORX);
    //printf("=  TARGET_UTIL                = %3.4f %%   \n", TARGET_UTIL);
    //printf("=  Theoretical L              = %f cust    \n", l_theory);
    //printf("=============================================================== \n");
    //printf("=  Mean service time          = %f sec      \n", Ts);
    //printf("=  Mean time between arrivals = %f sec      \n", Ta);
    //printf("=============================================================== \n");
    //printf("=  Simulation time  = %f sec      \n", time);
    //printf("=-------------------------------------------------------------= \n");
    //printf("=  L = %f cust     \n", l);
    //printf("=  W = %f sec      \n", w);
    //printf("=  U = %f %%       \n", 100.0 * u);
    //printf("=  C = %ld cust    \n", c);
    //printf("=  X = %f cust/sec \n", x);
    //printf("=============================================================== \n");

    // Summing of simulation time
    sum = sum + time;
  }

  // Output mean simulation time
  fprintf(st2, "&& Mean simulation time = %f \n", sum / MAX_ITER);
}

//=============================================================================
//=  Function to generate exponentially distributed RVs using inverse method  =
//=    - Input:  x (mean value of distribution)                               =
//=    - Output: Returns with exponential RV                                  =
//=============================================================================
double expntl(double x)
{
  double z;                 // Uniform random number from 0 to 1

  // Pull a uniform RV (0 < z < 1)
  do
  {

    z = rand_val(0);        // Jain's RNG
  }
  while ((z == 0) || (z == 1));

  return(-x * log(z));
}

//=========================================================================
//= Multiplicative LCG for generating uniform(0.0, 1.0) random numbers    =
//=   - x_n = 7^5*x_(n-1)mod(2^31 - 1)                                    =
//=   - With x seeded to 1 the 10000th x value should be 1043618065       =
//=   - From R. Jain, "The Art of Computer Systems Performance Analysis," =
//=     John Wiley & Sons, 1991. (Page 443, Figure 26.2)                  =
//=========================================================================
double rand_val(long seed)
{
  const long  a =      16807;  // Multiplier
  const long  m = 2147483647;  // Modulus
  const long  q =     127773;  // m div a
  const long  r =       2836;  // m mod a
  static long x;               // Seed value
  long        x_div_q;         // x divided by q
  long        x_mod_q;         // x modulo q
  long        x_new;           // New x value

  // Set the seed
  if (seed > 0) x = seed;

  // RNG using integer arithmetic
  x_div_q = x / q;
  x_mod_q = x % q;
  x_new = (a * x_mod_q) - (r * x_div_q);
  if (x_new > 0)
    x = x_new;
  else
    x = x_new + m;

  // Return a random value between 0.0 and 1.0
  return((double) x / m);
}
